# matrices.py - functions for matrices
#
# Copyright 2014 Jeffrey Finkelstein.
#
# This file is part of fraciso.
#
# fraciso is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# fraciso is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# fraciso.  If not, see <http://www.gnu.org/licenses/>.
"""Functions that operate on matrices."""
from __future__ import division  # Just in case this is run on Python 2.X

from itertools import permutations
from itertools import product

import numpy as np

#: The data type for each entry in matrices created by functions in this
#: module.
DTYPE = np.int32


def permutation_matrices(n):
    """Returns an iterator over all possible permutation matrices of size `n`.

    """
    return (np.mat(P) for P in permutations(np.identity(n, dtype=DTYPE)))


def permutation_to_matrix(permutation):
    """Returns a permutation matrix corresponding to the permutation given by
    the specified dictionary.

    """
    p = [v for k, v in sorted(permutation.items())]
    return np.identity(len(p), dtype=DTYPE)[p, :]


def random_permutation_matrix(n, seed=None):
    """Returns a random `n` by `n` permutation matrix.

    If `seed` is specified, it is used to specify the random seed for the
    pseudorandom number generator.

    """
    np.random.seed(seed)
    p = np.random.permutation(n)
    # Only need to permute the rows, since this is the identity matrix.
    return np.identity(n, dtype=DTYPE)[p, :]


def is_doubly_stochastic(matrix):
    """Returns ``True`` if and only if the specified NumPy matrix is doubly
    stochastic.

    A matrix is **doubly stochastic** if the sum of the entries in each row is
    1 and the sum of the entries in each column is 1.

    """
    rows, columns = matrix.shape
    ones_m = np.mat(np.ones(columns)).T
    ones_n = np.mat(np.ones(rows)).T
    return (np.all(matrix * ones_n == ones_m)
            and np.all(ones_m.T * matrix == ones_n))


def dictionary_to_permutation(permutation):
    """Converts the specified permutation given as a dictionary to the
    corresponding permutation matrix.

    `permutation` must be a dictionary representing a permutation on the first
    **n** nonnegative integers, where **n** is the length of the dictionary.

    """
    n = len(permutation)
    matrix = np.mat(np.zeros((n, n)))
    for k, v in permutation.items():
        matrix[k, v] = 1.
        #matrix[v, k] = 1.
    return matrix


def _swap(obj, i, j):
    """Swaps the elements at index `i` and index `j` in `obj`.

    If `obj` has a ``copy()`` method, ``obj[j]`` gets a copy of ``obj[i]``.

    """
    temp = obj[i]
    if hasattr(temp, 'copy'):
        temp = temp.copy()
    obj[i] = obj[j]
    obj[j] = temp
    return obj


def _index_of_nonzero(column):
    """Returns the index of the first nonzero element in the specified array.

    """
    nonzero_entries = np.flatnonzero(np.asarray(column))
    if len(nonzero_entries) == 0:
        return -1
    return nonzero_entries[0]


def to_row_echelon(A):
    """Returns the row echelon form of `A` as computed by Gaussian elimination.

    This function returns a two-tuple. The first element is the row echelon
    form of A (possibly including permutations of the rows). The second element
    is the dictionary indicating the permutation applied to the matrix during
    the computation of the row echelons form. The dictionary includes a mapping
    from **i** to **j** if row **i** was swapped with row **j**.

    This function returns a new matrix; it does not modify `A`.

    """
    A = A.copy()
    rows, columns = A.shape
    # permutation stores the pivots that were performed: row i of the original
    # matrix was interchanged with row permutation[i]
    permutation = {x: x for x in range(rows)}
    # p is the current pivot column.
    p = 0
    # Iterate over each row (except the last one).
    for i in range(rows - 1):
        # If the entry at the current expected pivot point is zero, find the
        # next non-zero entry in this column and swap it.
        while A[i, p] == 0:
            nonzero = _index_of_nonzero(A[i:, p])
            # If there is some non-zero element, swap its row with the current
            # row and break out of this loop to move on to the elimination.
            if nonzero != -1:
                # nonzero is the index relative to row i, so we add nonzero to
                # i to determine the absolute row number in the matrix A.
                _swap(A, i, i + nonzero)
                _swap(permutation, i, i + nonzero)
                break
            # Otherwise, all elements in the current pivot column are zeros,
            # so increment the pivot column to try and find a usable column.
            p += 1
            # If we have exhausted all possible columns, this means the rest of
            # the matrix is zeros.
            if p == columns:
                return A, permutation
        # TODO this could be done more succinctly using a list comprehension
        #
        # Iterate over each row below the current one.
        for j in range(i + 1, rows):
            # k is the scalar by which we multiply the row before adding it.
            k = -A[j, p] / A[i, p]
            A[j] += k * A[i]
        # Increment the pivot column when we move to the next row.
        p += 1
        if p == columns:
            return A, permutation
    return A, permutation


def _sequences_of_ones(n, k):
    """Returns a list of all binary sequences of length `n` with `k` ones."""
    # There is exactly one sequence of length zero: the empty sequence.
    if n <= 0:
        return [[]]
    # There is exactly one sequence with no ones: the all-zeros sequence.
    if k <= 0:
        return [[0 for x in range(n)]]
    # There is exactly one sequence with all ones: the all-ones sequence.
    if k >= n:
        return [[1 for x in range(n)]]
    # In the recursive case, there are two possibilities: prepend a 0 to the
    # sequence or prepend a 1 to the sequence.
    return [[0] + seq for seq in _sequences_of_ones(n - 1, k)] \
        + [[1] + seq for seq in _sequences_of_ones(n - 1, k - 1)]


def _matrices_with_row_sums(m, n, d):
    """Returns an iterator over all m by n matrices whose rows all sum to d.

    This function is dangerous! It creates a combinatorial explosion of
    matrices, so don't try to store all of the matrices in memory at once!

    """
    return product(_sequences_of_ones(n, d), repeat=m)


def _row_sum(matrix, m1, n1, m2, n2):
    """Computes the sum of the first row of the submatrix starting at (m1, n1)
    and ending at (m2, n2).

    `m2` must be greater than `m1` and `n2` must be greater than `n1`.

    """
    return sum(matrix[m1:m2, n1:n2][0])



# isomorphism.py - algorithms for checking fractional isomorphism of graphs
#
# Copyright 2014 Jeffrey Finkelstein.
#
# This file is part of fraciso.
#
# fraciso is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# fraciso is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# fraciso.  If not, see <http://www.gnu.org/licenses/>.
"""Algorithms for determining if two graphs are fractionally isomorphic."""
import itertools

from networkx import get_node_attributes
from networkx import Graph
from networkx.convert import from_numpy_matrix
from networkx.convert import to_numpy_matrix
from networkx.generators.bipartite import bipartite_configuration_model
from networkx.generators.random_graphs import random_regular_graph
import numpy as np
from scipy.linalg import block_diag

from fraciso.linprog import cvxopt_available
from fraciso.linprog import ecos_available
from fraciso.linprog import fraciso_using_lp
from fraciso.linprog import LP_METHODS
from fraciso.linprog import pulp_available
from fraciso.matrices import dictionary_to_permutation
from fraciso.matrices import is_doubly_stochastic
from fraciso.matrices import random_permutation_matrix
from fraciso.matrices import permutation_to_matrix
from fraciso.matrices import to_row_echelon
from fraciso.partitions import are_common_partitions
from fraciso.partitions import coarsest_equitable_partition
from fraciso.partitions import partition_parameters


def _cep_solver(G, H):
    """Solves the fractional graph isomorphism problem by comparing the
    coarsest equitable partitions of G and H.

    Returns a two-tuple whose left element is a Boolean representing whether
    the two graphs have a common coarsest partition and whose right element is
    the pair of partitions of the given graphs (in the same order as the inputs
    are given). If the left element is ``False``, the right element is
    ``None``, since there is no witness that the graphs are fractionally
    isomorphic.

    """
    partition1 = coarsest_equitable_partition(G)
    partition2 = coarsest_equitable_partition(H)
    result = are_common_partitions(G, partition1, H, partition2)
    witness = (partition1, partition2) if result else None
    return result, witness


def are_fractionally_isomorphic(G, H, algorithm='cep'):
    """Solves the fractional graph isomorphism problem and provides a witness,
    if any, that the two graphs are fractionally isomorphism.

    Two graphs are **fractionally isomorphic** if they share a common coarsest
    equitable partition. Equivalently, graphs with adjacency matrices A and B
    are fractionally isomorphic if there is a doubly stochastic matrix such
    that AS = SB.

    `G` and `H` must be instances of :class:`networkx.Graph`.

    `algorithm` can be either ``'cep'`` (the default) or ``'lp.LPSOLVER'``,
    where ``LPSOLVER`` is one of the recognized linear programming solvers
    (``ecos``, ``cvxopt``, etc.). The former denotes the coarsest equitable
    partition algorithm, which finds the coarsest equitable partition in either
    graph and checks that they are equivalent. The latter denotes one of the
    linear programming algorithms, which solves the linear programming
    formulation of the fractional graph isomorphism problem as specified above.

    The available linear programming solvers can be found in
    :data:`fraciso.linprog.LP_METHODS`.

    This function returns a two-tuple. The left element is a Boolean
    representing whether the two graphs are fractionally isomorphic. The right
    element is a witness that the two graphs are fractionally isomorphic, or
    ``None`` if no such witness exists. The witness depends on the algorithm

    If a specified linear programming library is known but unavailable, this
    function raises a :exc:`RuntimeError`. If an unknown algorithm is
    specified, this function raises a :exc:`ValueError`.

    """
    if algorithm == 'cep':
        return _cep_solver(G, H)
    if algorithm.startswith('lp'):
        try:
            method = algorithm.split('.')[1]
        except:
            raise ValueError('Must provide algorithm like "lp.ecos" or'
                             ' "lp.cvxopt"; got "{}"'.format(algorithm))
        if method not in LP_METHODS:
            raise ValueError('Unknown linear programming method'
                             ' {}'.format(method))
        if method == 'ecos' and not ecos_available:
            raise RuntimeError('ecos unavailable; try "pip install ecos"')
        if method == 'cvxopt' and not cvxopt_available:
            raise RuntimeError('cvxopt unavailable; try "pip install cvxopt"')
        if method == 'pulp' and not pulp_available:
            raise RuntimeError('pulp unavailable; try "pip install pulp"')
        return fraciso_using_lp(G, H, method)
    raise ValueError('Unknown algorithm: {}'.format(algorithm))


def _random_biregular_graph(num_left_vertices, num_right_vertices, left_degree,
                            right_degree, half_only=False, seed=None):
    """Returns the adjacency matrix of a random biregular graph with the
    specified number of left and right vertices, and the specified left and
    right degree.

    The returned adjacency matrix is a NumPy array (not a list of lists, not a
    NumPy matrix). If **L** and **R** are the number of left and right vertices
    respectively and **d** and **e** are the left and right degree
    respectively, then the returned adjacency matrix has the form::

         _      _
        |  0   B |
        |_B^T  0_|

    where B is an L by R matrix in which each row sums to **d** and each column
    sums to **e**. In this form, the first L nonnegative integers represent the
    left vertices and the following R integers represent the right vertices.

    If `half_only` is ``True``, this function returns only the submatrix B.

    If `seed` is specified, it must be an integer provided as the seed to the
    pseudorandom number generator used to generate the graph.

    .. note::

       Although Networkx includes a function for generating random bipartite
       graphs
       (:func:`networkx.generators.bipartite.bipartite_configuration_model`),
       their adjacency matrices do not necessarily have this block form.

    """
    # Rename some variables for brevity.
    L, R = num_left_vertices, num_right_vertices
    n = L + R
    # Generate a random graph with the appropriate degree sequence.
    left_sequence = [left_degree for x in range(L)]
    right_sequence = [right_degree for x in range(R)]
    # Need to use `create_using=Graph()` or else networkx will create a
    # multigraph.
    graph = bipartite_configuration_model(left_sequence, right_sequence,
                                          create_using=Graph(), seed=seed)
    # Find the nodes that are in the left and right sets. The `bipartite`
    # attribute specifies which set each vertex is in.
    left_or_right = get_node_attributes(graph, 'bipartite')
    left_nodes = (v for v, side in left_or_right.items() if side == 0)
    right_nodes = (v for v, side in left_or_right.items() if side == 1)
    all_nodes = itertools.chain(left_nodes, right_nodes)
    # Determine the permutation that moves all the left nodes to the top rows
    # of the matrix and all the right nodes to the bottom rows of the
    # matrix. The permutation maps row number to vertex that should be in that
    # row.
    permutation = {i: v for i, v in enumerate(all_nodes)}
    P = permutation_to_matrix(permutation)
    # Apply the permutation matrix to the adjacency matrix.
    M = to_numpy_matrix(graph)
    result = P * M
    # If `half_only` is True, only return the submatrix block consisting of the
    # first L rows and the last R columns.
    return result[:L, -R:] if half_only else result


def _random_graph_from_parameters(vertices_per_block, block_neighbors,
                                  seed=None):
    """Returns a random graph that satisfies the given parameters.

    `vertices_per_block` and `block_neighbors` are the matrices returned by
    :func:`~fraciso.partitions.partition_parameters`.

    If `seed` is specified, it must be an integer provided as the seed to the
    pseudorandom number generator used to generate the graph.

    """
    # TODO there is an alternate way to implement this function: create a
    # random regular networkx.Graph object for each block of the partition,
    # create a random biregular Graph object between blocks of the partition,
    # then compute the union of the two graphs.
    #
    # Rename some variables for the sake of brevity.
    n, D = np.asarray(vertices_per_block), np.asarray(block_neighbors)
    # p is the number of blocks
    p = len(n)
    mat = to_numpy_matrix
    rr = lambda d, s: random_regular_graph(d, s, seed=seed)
    rb = lambda L, R, d, e:  _random_biregular_graph(L, R, d, e, True, seed)
    # Create a block diagonal matrix that has the regular graphs corresponding
    # to the blocks of the partition along its diagonal.
    regular_graphs = block_diag(*(mat(rr(d, s))
                                  for s, d in zip(n, D.diagonal())))
    # Create a block strict upper triangular matrix containing the upper-right
    # blocks of the bipartite adjacency matrices.
    #
    # First, we create a list containing only the blocks necessary.
    blocks = [[rb(n[i], n[j], D[i, j], D[j, i]) for j in range(i + 1, p)]
              for i in range(p - 1)]
    # Next, we pad the lower triangular entries with blocks of zeros. (We also
    # need to add an extra block row of all zeros.) At this point, `padded` is
    # a square list of lists.
    padded = [[np.zeros((n[i], n[j])) for j in range(p - len(row))] + row
              for i, row in enumerate(blocks)]
    padded.append([np.zeros((n[-1], n[i])) for i in range(p)])
    # To get the block strict upper triangular matrix, we concatenate the block
    # matrices in each row.
    biregular_graphs = np.vstack(np.hstack(row) for row in padded)
    # Finally, we add the regular graphs on the diagonaly, the upper biregular
    # graphs, and the transpose of the upper biregular graphs in order to get a
    # graph that has the specified parameters.
    adjacency_matrix = regular_graphs + biregular_graphs + biregular_graphs.T
    return from_numpy_matrix(adjacency_matrix)


def random_fractionally_isomorphic_graph(graph, seed=None):
    """Returns a random graph that is fractionally isomorphic to the specified
    graph.

    This function may return the same graph.

    """
    # Get the parameters for the coarsest equitable partition of the graph; n
    # is the vector containing number of vertices in each block of the
    # partition and D_ij is the matrix containing the degree of any vertex in
    # block i relative to block j. p is the number of blocks in the partition.
    partition = coarsest_equitable_partition(graph)
    n, D = partition_parameters(graph, partition, as_matrices=True)
    p = len(n)
    # Permuting the parameters in an arbitrary fashion may produce parameters
    # for which there is no possible graph. For now, just stick with a random
    # graph on the same parameters.
    #
    ## Choose a random permutation matrix and permute the parameters.
    #P = random_permutation_matrix(p, seed)
    #n, D = P * n, P * D
    #
    # Get a random graph with those parameters.
    return _random_graph_from_parameters(n, D, seed)


def random_fractionally_isomorphic_graphs(graph, number=None, seed=None):
    """Returns an iterator that generates random graphs that are fractionally
    isomorphic to the specified graph.

    If `number` is specified, the iterator only generates that number of graphs
    before it terminates. If it is not specified, it will generate graphs
    forever!

    If `seed` is specified, it must be an integer provided as the seed to the
    pseudorandom number generator used to generate the graph.

    This function may return the same graph.

    """
    # Get the parameters for the coarsest equitable partition of the graph; n
    # is the vector containing number of vertices in each block of the
    # partition and D_ij is the matrix containing the degree of any vertex in
    # block i relative to block j. p is the number of blocks in the partition.
    partition = coarsest_equitable_partition(graph)
    n, D = partition_parameters(graph, partition, as_matrices=True)
    p = len(n)
    # This code is almost exactly the same as itertools.repeat(), except that
    # we execute the _random_graph_from_parameters() function every time
    # through the loop.
    if times is None:
        while True:
            # Permuting the parameters in an arbitrary fashion may produce
            # parameters for which there is no possible graph. For now, just
            # stick with a random graph on the same parameters.
            #
            ## Choose a random permutation matrix and permute the parameters.
            #P = random_permutation_matrix(p, seed)
            #n_prime, D_prime = P * n, P * D
            #
            # Get a random graph with those parameters.
            yield _random_graph_from_parameters(n, D, seed)
    else:
        for i in range(times):
            # (See the note on permutations above.)
            #
            # Get a random graph with those parameters.
            yield _random_graph_from_parameters(n, D, seed)


def fractionally_isomorphic_graphs(graph):
    """Returns an iterator over all graphs that are fractionally isomorphic to
    the specified graph, including the graph itself.

    `graph` must be an instance of :class:`networkx.Graph`.

    .. warning::

       This function is currently not implemented, since it is very difficult
       to write correct code that enumerates each graph that is fractionally
       isomorphic to a given graph (that is, it seems that the code will be
       very complicated).

    """
    raise NotImplementedError


def verify_isomorphism(G, H, S):
    """Returns ``True`` if and only if the doubly stochastic matrix `S` is a
    correct witness that `G` and `H` are fractionally isomorphic.

    `G` and `H` are instances of :class:`networkx.Graph`. `S` is a Numpy
    matrix.

    If **A** and **B** are the adjacency matrices of `G` and `H`, respectively,
    then this function checks that **AS = SB**, and that `S` is nonnegative and
    doubly stochastic.

    """
    A = to_numpy_matrix(G)
    B = to_numpy_matrix(H)
    return (np.all(S >= 0) and is_doubly_stochastic(S)
            and np.array_equal(A * S, S * B))



# partitions.py - functions for modifying partitions of a graph
#
# Copyright 2014 Jeffrey Finkelstein.
#
# This file is part of fraciso.
#
# fraciso is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# fraciso is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# fraciso.  If not, see <http://www.gnu.org/licenses/>.
"""Functions for computing and modifying partitions of a graph.

A **partition** of a graph is a collection of sets of vertices that are
pairwise disjoint and whose union equals the set of all vertices in the graph.

"""
from itertools import accumulate
from itertools import chain
from itertools import combinations
from itertools import product
from itertools import groupby

import numpy as np

from fraciso.matrices import permutation_matrices


def peek(s):
    """Returns but does not remove an arbitrary element from the specified set.

    """
    return next(iter(s))


def union(*sets):
    """Returns the union of all sets given as positional arguments."""
    # The last argument to reduce is the initializer used in the case of an
    # empty sequence of sets.
    #
    # This is essentially the same as ``sets[0].union(*sets[1:])``, but it
    # works even if `sets` is not a list and if `sets` has only one element.
    return reduce(lambda S, T: S | T, sets, {})


def are_pairwise_disjoint(*sets):
    """Returns ``True`` if and only if each pair of sets is disjoint."""
    return all(S.isdisjoint(T) for S, T in combinations(sets, 2))


def is_valid_partition(graph, partition):
    """Returns ``True`` if and only if the specified partition is a valid
    partition of the vertices of `graph`.

    `graph` is an instance of :class:`networkx.Graph`.

    `partition` is a set of sets of vertices from the graph.

    A partition is valid if the set of vertices of the graph equals the
    disjoint union of the blocks in the partition.

    """
    return (are_pairwise_disjoint(*partition)
            and union(*partition) == set(graph.nodes())
            and all(len(block) > 0 for block in partition))


def degree(graph, v, block=None):
    """Returns the number of neighbors of vertex `v` in the specified graph.

    `graph` is an instance of :class:`networkx.Graph`.

    If `block` is a set of vertices in the given graph, this function returns
    only the number of neighbors in the specified block. Otherwise, it simply
    returns the degree of `v` in `graph`.

    """
    if block is None:
        block = set(graph)
    subgraph = graph.subgraph(block | {v})
    return subgraph.degree(v)


def is_block_equitable(graph, partition, block):
    """Returns ``True`` if and only if the specified block of the given
    partition is equitable.

    `graph` must be an instance of :class:`networkx.Graph`.

    `partition` is a set of sets. Each inner set, a **block**, is a set of
    vertices in the graph.

    `block` must be an element of `partition`. This block will be checked for
    equitability.

    A block is **equitable** if for each two vertices v and w in the block, the
    number of neighbors of v in block B equals the number of neighbors of w in
    block B, for all blocks B.

    """
    return all(degree(graph, v, B) == degree(graph, w, B)
               for v, w in product(block, block) for B in partition)


def is_partition_equitable(graph, partition):
    """Returns ``True`` if and only if the specified partition is an equitable
    partition of the graph.

    `graph` must be an instance of :class:`networkx.Graph`.

    `partition` is a set of sets. Each inner set, a **block**, is a set of
    vertices in the graph.

    A partition is **equitable** if for each block P1 and for each two vertices
    v and w in P1, the number of neighbors of v in block P2 equals the number
    of neighbors of w in block P2, for all blocks P2.

    """
    return all(is_block_equitable(graph, partition, block)
               for block in partition)


def _adapt(graph, partition):
    """Adapt the specified equitable partition into the coarsest equitable one.

    `graph` must be an instance of :class:`networkx.Graph`.

    `partition` is a set of blocks, each of which is a :class:`frozenset` of
    vertices from the graph.

    """
    # First, for each vertex, compute the number of neighbors of that vertex in
    # each block (including its own).
    block_neighbors = {v: {block: degree(graph, v, block)
                           for block in partition}
                       for v in graph}
    # Second, partition all vertices into new blocks consisting of vertices
    # that have the same dictionary of block neighbors as computed above.
    #
    # We do this by first sorting all the vertices according to their block
    # neighbors dictionary, then grouping together the vertices that have the
    # same block neighbors dictionary. However, since dictionaries cannot be
    # used as sort keys (because they are not comparable), we use the sorted
    # list of key/value pairs. Even more annoying, the keys in these key/value
    # pairs are themselves sets, which must be converted to sorted lists
    # (because the default comparison on sets in Python is the "is subset"
    # comparison).
    #
    # This is quite an inefficient algorithm for doing this, but the reward is
    # readability to match the mathematical description of the algorithm more
    # closely.
    key_func = lambda w: sorted((sorted(k), v)
                                for k, v in block_neighbors[w].items())
    vertices = sorted(block_neighbors, key=key_func)
    # Next, we simply group the vertices according to ones that have the same
    # block neighbors dictionary.
    new_partition = {frozenset(g) for k, g in groupby(vertices, key_func)}

    # Do a sanity check...
    #assert is_valid_partition(new_partition)

    # If this process produced exactly the same partition, then we have reached
    # the base case of the recursion.
    if new_partition == partition:
        return partition
    return _adapt(graph, new_partition)


def coarsest_equitable_partition(graph):
    """Returns the coarsest equitable partition of the specified graph.

    `graph` must be an instance of :class:`networkx.Graph`.

    A partition is a set of sets. Each of the inner sets is a
    :class:`frozenset` of vertices in the graph.

    A partition is **equitable** if for each block P1 and for each two vertices
    v and w in P1, the number of neighbors of v in block P2 equals the number
    of neighbors of w in block P2, for all blocks P2.

    A partition P1 is coarser than partition P2 if each block of P2 is a subset
    of a block of P1. A partition is a coarsest equitable partition if it is
    equitable and no coarser partition is equitable.

    """
    # Start with the partition that consists of a single block containing the
    # entire vertex set. Recursively adapt the partition until no further
    # adaptations are possible. This is guaranteed to be the coarsest equitable
    # partition.
    return _adapt(graph, {frozenset(graph)})


def partition_parameters(graph, partition, as_matrices=False):
    """Returns the parameters of the given partition.

    `graph` must be an instance of :class:`networkx.Graph`.

    `partition` must be a valid equitable partition of the specified graph.

    The parameters of the graph are two dictionaries. The first dictionary is a
    mapping from a block of the partition to the number of vertices in that
    block. The second dictionary is a two-dimensional dictionary. Both keys are
    blocks of the partition. The entry indexed by blocks **i** and **j** is the
    number of neighbors in block **j** of any fixed vertex in block **i** (the
    particular vertex doesn't matter, since the partition is equitable).

    If `as_matrices` is ``True``, then the parameters will be converted to
    NumPy matrices.

    """
    vertices_per_block = {block: len(block) for block in partition}
    block_neighbors = {b_i: {b_j: degree(graph, peek(b_i), b_j)
                             for b_j in partition}
                       for b_i in partition}
    if not as_matrices:
        return vertices_per_block, block_neighbors
    return _as_vector(vertices_per_block), _as_matrix(block_neighbors)


def partition_to_permutation(graph, partition):
    """Converts the specified partition of the graph to a permutation matrix,
    representing the permutation of the adjacency matrix of the graph that
    places rows into contiguous blocks representing blocks of the partition.

    `graph` must be an instance of :class:`networkx.Graph`. Furthermore, the
    **n** vertices of the graph must be the first **n** nonnegative integers.

    `partition` must be a valid partition of `graph`.

    """
    n = len(graph)
    # Sort the blocks of the partition according to lexicographic order, and
    # sort the set of blocks as well so that the output is well-defined.
    sorted_partition = sorted(sorted(block) for block in partition)
    extents = list(accumulate(len(block) for block in sorted_partition))
    # Flatten the partition so that we have simply a sequence of vertices
    # representing a permutation.
    permutation = chain(*sorted_partition)
    # Create the permutation matrix from the permutation.
    matrix = np.matrix([[1 if j == v else 0 for j in range(n)]
                        for v in permutation])
    return matrix, extents


def lexicographic_blocks(dictionary):
    """Sorts the key/value pairs of `dictionary` according to the lexicographic
    order of the keys.

    This is useful if your keys are instances of :class:`set` and you wish them
    to be ordered according to their list representation instead of ordered
    according to inclusion.

    """
    return sorted(dictionary.items(), key=lambda item: sorted(item[0]))


def _as_vector(sizes):
    """Converts the given dictionary to a vector.

    The dictionary `sizes` maps blocks of a partition to number of vertices in
    that partition. Each element of the returned list is the number of vertices
    in a block of the partition. The list is sorted according to the
    lexicographic ordering of the blocks.

    The returned

    """
    return np.mat([size for block, size in lexicographic_blocks(sizes)]).T


def _as_matrix(block_neighbors):
    """Converts the given dictionary to a matrix.

    The two-dimensional dictionary `block_neighbors` is indexed in each
    dimension by blocks of a partition. The entry at row **i**, column **j** is
    an integer representing the number of neighbors in block **j** of any fixed
    vertex in block **i**. Each row and each column of the matrix is sorted
    according to lexicographic order of the blocks.

    """
    return np.mat([[num for block_j, num in lexicographic_blocks(d)]
                   for block_i, d in lexicographic_blocks(block_neighbors)])


def are_common_partitions(graph1, partition1, graph2, partition2):
    """Returns ``True`` if and only if the two partitions have the same
    parameters (or a permutation of the same parameters).

    `graph1` and `graph2` must be instances of :class:`networkx.Graph`.

    `partition1` and `partition2` must be valid equitable partitions of the
    corresponding graphs.

    The partition parameters are described in the documentation for
    :func:`partition_parameters`. The parameters include one vector and one
    matrix. This function returns ``True`` exactly when there is some
    permutation such that both the vector and the matrix of `partition1` equal
    the vector and the matrix of `partition2`.

    """
    # By specifying `as_matrices=True`, `sizes` becomes a vector of length p
    # and `block_neighbors` a matrix of size p by p, where p is the number of
    # blocks in the partitions. The innermost entries of both should be
    # integers.
    sizes1, neighbors1 = partition_parameters(graph1, partition1,
                                              as_matrices=True)
    sizes2, neighbors2 = partition_parameters(graph2, partition2,
                                              as_matrices=True)
    # Return true if there is any permutation that makes these two pairs equal.
    match = lambda P: (np.array_equal(sizes1, P * sizes2)
                       and np.array_equal(neighbors1, P * neighbors2))
    return any(match(P) for P in permutation_matrices(len(sizes1)))



# linprog.py - interfaces to linear programming solvers
#
# Copyright 2014 Jeffrey Finkelstein.
#
# This file is part of fraciso.
#
# fraciso is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# fraciso is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# fraciso.  If not, see <http://www.gnu.org/licenses/>.
"""Interfaces to linear programming solvers for use in determining whether two
graphs are fractionally isomorphic.

"""
try:
    from cvxopt import matrix as cvx_matrix
    from cvxopt import solvers as cvx_solvers
    cvxopt_available = True
except ImportError:
    cvxopt_available = False
try:
    import ecos
    from scipy.sparse import csc_matrix
    ecos_available = True
except ImportError:
    ecos_available = False
from networkx.convert import to_numpy_matrix
import numpy as np
try:
    #from pulp import COIN
    from pulp import GLPK
    from pulp import LpMinimize
    from pulp import LpProblem
    from pulp import LpStatusOptimal
    from pulp import lpSum
    from pulp import LpVariable
    pulp_available = True
except ImportError:
    pulp_available = False

from fraciso.matrices import dictionary_to_permutation
from fraciso.matrices import to_row_echelon


#: Known linear programming backends.
LP_METHODS = ('ecos', 'cvxopt', 'pulp')


def _pulp_dot_product(costs, variables):
    """Returns the PuLP-specific "dot product" of the given iterables.

    `costs` is an iterable of numbers specifying the weight or cost of the
    corresponding variable.

    `variables` is an iterable of :class:`LpVariable` instances.

    `costs` and `variables` must have the same length.

    """
    return lpSum(a * b for a, b in zip(costs, variables))


def pulp_solver(G, h, A, b, c, n):
    # First, create a variable for each of the columsn of G and A.
    #
    # pre-condition: G and A have the same number of columns.
    #
    # The second argument specifies a lower bound for the variable, so we can
    # safely ignore the inequality constraints given by G and h.
    variables = [LpVariable('s{}'.format(i), 0) for i in range(G.shape[1])]
    # LpVariable has a second argument that allows you to specify a lower bound
    # for the variable (for example, x1 >= 0). We don't specify nonnegativity
    # here, because it is already specified by the inequality constraints G and
    # h.
    #variables = [LpVariable('s{}'.format(i)) for i in range(G.shape[1])]
    # Next, create a problem context object and add the objective function c to
    # it. The first object added to LpProblem is implicitly interpreted as the
    # objective function.
    problem = LpProblem('fraciso', LpMinimize)
    # The np.dot() function doesn't like mixing numbers and LpVariable objects,
    # so we compute the dot product ourselves.
    #
    #problem += np.dot(variables, c), 'Dummy objective function'
    problem += _pulp_dot_product(c, variables), 'Dummy objective function'
    # Add each equality constraint to the problem context.
    for i, (row, b_value) in enumerate(zip(A, b)):
        #problem += np.dot(row, variables), 'Constraint {}'.format(i)
        # Convert the row to a list so pulp has an easier time dealing with it.
        row_as_list = np.asarray(row).flatten().tolist()
        dot_product = _pulp_dot_product(row_as_list, variables)
        problem += dot_product == b_value, 'Constraint {}'.format(i)
    solver_backend = GLPK()
    #solver_backend = COIN()
    problem.solve(solver_backend)
    if problem.status == LpStatusOptimal:
        # PuLP is silly and sorts the variables by name before returning them,
        # so we need to re-sort them in numerical order.
        solution = [s.varValue for s in sorted(problem.variables(),
                                               key=lambda s: int(s.name[1:]))]
        return True, solution
    # TODO status could be unknown here, but we're currently ignoring that
    return False, None


def ecos_solver(G, h, A, b, c, n):
    # In order to use ecos, we must provide sparse matrices
    A = csc_matrix(A)
    G = csc_matrix(G)
    dims = dict(l=n ** 2, q=[])
    solution = ecos.solve(c, G, h, dims, A, b)
    if solution['info']['exitFlag'] == 0:
        return True, solution['x']
    # TODO status could be unknown here, but we're currently ignoring that
    return False, None


def cvxopt_solver(G, h, A, b, c, n):
    # cvxopt doesn't allow redundant constraints in the linear program Ax = b,
    # so we need to do some preprocessing to find and remove any linearly
    # dependent rows in the augmented matrix [A | b].
    #
    # First we do Gaussian elimination to put the augmented matrix into row
    # echelon (reduced row echelon form is not necessary). Since b comes as a
    # numpy array (that is, a row vector), we need to convert it to a numpy
    # matrix before transposing it (that is, to a column vector).
    b = np.mat(b).T
    A_b = np.hstack((A, b))
    A_b, permutation = to_row_echelon(np.hstack((A, b)))
    # Next, we apply the inverse of the permutation applied to compute the row
    # echelon form.
    P = dictionary_to_permutation(permutation)
    A_b = P.I * A_b
    # Trim any rows that are all zeros. The call to np.any returns an array of
    # Booleans that correspond to whether a row in A_b is all zeros. Indexing
    # A_b by an array of Booleans acts as a selector. We need to use np.asarray
    # in order for indexing to work, since it expects a row vector instead of a
    # column vector.
    A_b = A_b[np.any(np.asarray(A_b) != 0, axis=1)]
    # Split the augmented matrix back into a matrix and a vector.
    A, b = A_b[:, :-1], A_b[:, -1]
    # Apply the linear programming solver; cvxopt requires that these are all
    # of a special type of cvx-specific matrix.
    G, h, A, b, c = (cvx_matrix(M) for M in (G, h, A, b, c))
    solution = cvx_solvers.lp(c, G, h, A, b)
    if solution['status'] == 'optimal':
        return True, solution['x']
    # TODO status could be 'unknown' here, but we're currently ignoring that
    return False, None


def fraciso_using_lp(graph1, graph2, method='ecos'):
    """Solves the fractional graph isomorphism problem by solving the
    corresponding linear program.

    `graph1` and `graph2` are instances of :class:`networkx.Graph`.

    `method` is a string indicating which linear programming library to use.
    Possible values are ``'ecos'``_ (the default), ``'cvxopt'``_, and
    ``'pulp'``_ (see also :data:`LP_METHODS`).

    If M and N are the adjacency matrices of the two graphs, then deciding if G
    is fractionally isomorphic to H is equivalent to deciding if the following
    linear program is feasible.

        M S = S N
        S 1 = 1
        1^T S = 1
        S >= 0

    Matrices separated by a single space denotes matrix multiplication. The
    ``1`` denotes the all ones vector. The first equality requires that there
    is a fuzzy alignment between the vertices of G and the vertices of H. The
    second requires that S is row-stochastic. The third requires that S is
    column-stochastic. The last is a shorthand for component-wise inequality,
    and requires that S is real but non-negative.

    This can be translated into a more standardized form of linear programming
    constraints of the form

        G x <= h
        A x  = b

    by setting the variables to be

        x = [s_11, s_12, ..., s_1n, s_21, ..., s_2n, ..., s_n1, ...s_nn]^T

    and setting the matrices as follows. (Below, ``(x)`` denotes the tensor
    product, also known as the "Kronecker product" [this is how it is known to
    NumPy].)

        G = -I (of dimension n^2 by n^2)
        h = 0  (of dimension n^2 by 1)
            __                       __
        A = |       (I (x) 1)^T       |  (of dimension 2n + n^2 by n^2)
            |       (1 (x) I)^T       |
            |_(M (x) I) - (I (x) N^T)_|
            __ __
        b = | 1 |  (of dimension 2n + n^2 by 1)
            | 1 |
            |_0_|

    Since we require an objective function in order to have a well-defined
    instance of the linear programming problem, we choose an arbitrary
    objective function, say the all ones vector, since we care only about
    feasibility, not about optimality.

    .. _ecos: https://github.com/ifa-ethz/ecos
    .. _cvxopt: http://cvxopt.org/
    .. _pulp: https://code.google.com/p/pulp-or

    """
    # Sanity check
    #assert len(G) == len(H)
    # For the sake of brevity, let I(n) be the identity matrix of dimension n.
    I = np.identity
    # Get the number of vertices in the graph.
    n = len(graph1)
    # Let M and N be the adjacency matrices of G and H respectively.
    M = to_numpy_matrix(graph1)
    N = to_numpy_matrix(graph2)
    # Set G and h, the linear inequalities part of the linear program. We use
    # these only to require that each entry of S must be positive.
    G = -I(n ** 2)
    h = np.zeros(n ** 2)
    # Set A and b, the linear equalities part of the linear program. We use
    # these to require three things. First, that S is row-stochastic. Second,
    # that S is column-stochastic. Third, that M S = S N, that is, that there
    # is a fuzzy correlation between blocks of vertices in the graphs.
    #
    # To reflect these three requirements, A and b are both defined in three
    # blocks, as shown in the docstring. In NumPy, they would be defined as
    # follows.
    A_1 = np.kron(I(n), np.ones(n))
    A_2 = np.kron(np.ones(n), I(n))
    A_3 = np.kron(M, I(n)) - np.kron(I(n), N.T)
    A = np.vstack((A_1, A_2, A_3))
    b_1 = np.ones(n)
    b_2 = np.ones(n)
    b_3 = np.zeros(n ** 2)
    b = np.hstack((b_1, b_2, b_3))
    # Use a bogus objective function, since we care only if a feasible solution
    # to the linear program exists.
    c = np.ones(n ** 2)
    if method == 'ecos':
        solver = ecos_solver
    elif method == 'cvxopt':
        solver = cvxopt_solver
    elif method == 'pulp':
        solver = pulp_solver
    else:
        raise ValueError('Unknown linear programming method:'
                         ' {}'.format(method))
    is_feasible, solution = solver(G, h, A, b, c, n)
    # Convert the solution vector into the shape of the doubly stochastic
    # matrix S that witnesses the equality MS = SN, where M and N are the
    # adjacency matrices of the two graphs.
    S = np.array(solution).reshape(n, n) if is_feasible else None
    return is_feasible, S



# __init__.py - indicates that this directory is a Python package
#
# Copyright 2014 Jeffrey Finkelstein.
#
# This file is part of fraciso.
#
# fraciso is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# fraciso is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# fraciso.  If not, see <http://www.gnu.org/licenses/>.
from fraciso.isomorphism import are_fractionally_isomorphic
#from fraciso.isomorphism import fractionally_isomorphic_graphs
from fraciso.isomorphism import random_fractionally_isomorphic_graph
from fraciso.isomorphism import random_fractionally_isomorphic_graphs
from fraciso.isomorphism import verify_isomorphism
from fraciso.partitions import coarsest_equitable_partition



# setup.py - installation script for this package
#
# Copyright 2014 Jeffrey Finkelstein.
#
# This file is part of fraciso.
#
# fraciso is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# fraciso is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# fraciso.  If not, see <http://www.gnu.org/licenses/>.
"""Algorithms for fractional graph isomorphism."""
from setuptools import setup

#: Installation requirements.
requirements = ['networkx', 'numpy', 'scipy']


setup(
    author='Jeffrey Finkelstein',
    author_email='jeffrey.finkelstein@gmail.com',
    #classifiers=[],
    description='Algorithms for fractional graph isomorphism',
    download_url='https://github.com/jfinkels/fraciso',
    install_requires=requirements,
    include_package_data=True,
    #keywords=[],
    license='GNU GPLv3+',
    long_description=__doc__,
    name='fraciso',
    platforms='any',
    packages=['fraciso'],
    test_suite='nose.collector',
    tests_require=['nose'],
    url='https://github.com/jfinkels/fraciso',
    version='0.0.1-dev',
    zip_safe=False
)
